"""`StepImplementer` for the `report` step which builds and pushes and artifacts archive.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key           | Required? | Default | Description
----------------------------|-----------|---------|-------------
`organization`              | True      |         | Organization that this workflow is for. \
                                                    Used in archive name.
`application-name`          | True      |         | Application that this workflow is for. \
                                                    Used in archive name.
`service-name`              | True      |         | Service that this workflow is for. \
                                                    Used in archive name.
`version`                   | True      |         | Version of applicaiton service that this \
                                                    workflow is for. \
                                                    Used in archive name.
`results-archive-format`    | True      | 'zip'   | Archive format to use. For valid values see \
                                https://docs.python.org/3/library/shutil.html#shutil.make_archive
`results-archive-ignored-artifacts`  \
                            | False     | `['package-artifacts', 'image-tar-file']` | \
    Result artifacts to not include in this this result artifacts archive.
`results-archive-destination-url` \
                            | False     |         | URL to upload the results archive to.\
                                                    NOTE: TODO describe how URI is formed
                                                    <br /><br />\
                                                    Must start with `/` or `file://` to upload \
                                                    to local file path. \
                                                    <br /><br />\
                                                    Or must start with `http://` or `https://` to \
                                                    upload via a `PUT` to a remote location.
`results-archive-destination-username` \
                            | No        |         | Username to use when doing upload via http(s).
`results-archive-destination-password` \
                            | No        |         | Password to use when doing upload via http(s).

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key              | Description
---------------------------------|------------
`result-artifacts-archive`       | local path to the generated results artifacts archive.
`results-archive-uri`            | URI of the uploaded results archive.
`results-archive-upload-results` | Results of uploading the results archive to \
                                   the given destination.

"""

import json
import os
import pprint
import shutil

from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.utils.file import upload_file

DEFAULT_CONFIG = {
    'results-archive-format': 'zip',
    'results-archive-ignored-artifacts': [
        'package-artifacts',
        'image-tar-file'
    ]
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'organization',
    'application-name',
    'service-name',
    'version',
    'results-archive-format'
]
class ResultArtifactsArchive(StepImplementer):
    """`StepImplementer` for the `report` step which builds and pushes and artifacts archive.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        results_artifacts_archive = self.__create_archive()

        if results_artifacts_archive:
            results_artifacts_archive_value = results_artifacts_archive
        else:
            results_artifacts_archive_value = 'No result artifact values to archive.'

        step_result.add_artifact(
            name='result-artifacts-archive',
            value=results_artifacts_archive_value,
            description='Archive of all of the step result artifacts marked for archiving.'
        )

        # if destination URL specified, then upload the results archvie
        results_archive_destination_url = self.get_value('results-archive-destination-url')
        if results_artifacts_archive and results_archive_destination_url:
            org = self.get_value('organization')
            app = self.get_value('application-name')
            service = self.get_value('service-name')
            results_artifacts_archive_name = os.path.basename(results_artifacts_archive)
            results_archive_destination_uri = f"{results_archive_destination_url}/" \
                f"{org}/{app}/{service}/{results_artifacts_archive_name}"
            step_result.add_artifact(
                name='results-archive-uri',
                description='URI of the uploaded results archive.',
                value=results_archive_destination_uri
            )

            try:
                upload_result = upload_file(
                    file_path=results_artifacts_archive_value,
                    destination_uri=results_archive_destination_uri,
                    username=self.get_value('results-archive-destination-username'),
                    password=self.get_value('results-archive-destination-password')
                )
                step_result.add_artifact(
                    name='results-archive-upload-results',
                    description='Results of uploading the results archive ' \
                        'to the given destination.',
                    value=upload_result
                )
            except RuntimeError as error:
                step_result.success = False
                step_result.message = str(error)

        return step_result

    def __create_archive(self): # pylint: disable=too-many-branches,too-many-locals
        """Create the results artifact archive.

        Returns
        -------
        str
            Path to the created archive.
        """
        org = self.get_value('organization')
        app = self.get_value('application-name')
        service = self.get_value('service-name')
        version = self.get_value('version')
        result_archive_name = f"{org}-{app}-{service}-{version}"
        results_archive_root_dir_path = self.create_working_dir_sub_dir()

        ignored_artifacts = self.get_value('results-archive-ignored-artifacts')
        for previous_step_result in self.workflow_result.workflow_list:
            artifacts = previous_step_result.artifacts
            for artifact in artifacts.values():
                # only archive artifacts not on the ignore list
                if artifact.name not in ignored_artifacts:

                    # if value is path then copy the file(s) at path to archive location
                    # else create file with contents of value
                    if isinstance(artifact.value, str) and os.path.exists(artifact.value):
                        # create directory with name of artifact to put file(s) in
                        if previous_step_result.environment:
                            results_archive_artifact_dir_path = self.create_working_dir_sub_dir(
                                os.path.join(
                                    result_archive_name,
                                    previous_step_result.step_name,
                                    previous_step_result.sub_step_name,
                                    previous_step_result.environment,
                                    artifact.name
                                )
                            )
                        else:
                            results_archive_artifact_dir_path = self.create_working_dir_sub_dir(
                                os.path.join(
                                    result_archive_name,
                                    previous_step_result.step_name,
                                    previous_step_result.sub_step_name,
                                    artifact.name
                                )
                            )

                        if os.path.isfile(artifact.value):
                            shutil.copy(artifact.value, results_archive_artifact_dir_path)
                        elif os.path.isdir(artifact.value):
                            dest_path = os.path.join(
                                results_archive_artifact_dir_path,
                                os.path.basename(artifact.value)
                            )
                            shutil.copytree(artifact.value, dest_path)
                    else:
                        # create directory for step to put file with name of artifact in
                        if previous_step_result.environment:
                            results_archive_sub_step_dir_path = self.create_working_dir_sub_dir(
                                os.path.join(
                                    result_archive_name,
                                    previous_step_result.step_name,
                                    previous_step_result.sub_step_name,
                                    previous_step_result.environment
                                )
                            )
                        else:
                            results_archive_sub_step_dir_path = self.create_working_dir_sub_dir(
                                os.path.join(
                                    result_archive_name,
                                    previous_step_result.step_name,
                                    previous_step_result.sub_step_name
                                )
                            )
                        results_archive_artifact_file_path = os.path.join(
                            results_archive_sub_step_dir_path,
                            artifact.name
                        )

                        with open(results_archive_artifact_file_path, 'w', encoding='utf-8') \
                                as results_archive_artifact_file:
                            # format the data to print to file based on its type
                            if isinstance(artifact.value, str):
                                formated_artifact_value = artifact.value
                            elif isinstance(artifact.value, (dict, list)):
                                formated_artifact_value = json.dumps(
                                    artifact.value,
                                    indent=4
                                )
                            else:
                                printer = pprint.PrettyPrinter()
                                formated_artifact_value = printer.pformat(artifact.value)

                            # write data to file
                            results_archive_artifact_file.write(formated_artifact_value)

        # make the archive if there was anyting to archive
        archive_base_name = os.path.join(results_archive_root_dir_path, result_archive_name)
        if os.path.exists(archive_base_name):
            results_artifacts_archive = shutil.make_archive(
                base_name=archive_base_name,
                format=self.get_value('results-archive-format'),
                root_dir=results_archive_root_dir_path,
                base_dir=result_archive_name
            )
        else:
            results_artifacts_archive = None

        return results_artifacts_archive



"""`StepImplementer` for the `sign-container-image` step using Podman to create an image signature.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key      | Required? | Default | Description
-----------------------|-----------|---------|-------------
`signer-pgp-private-key` \
                       | Yes       |         | PGP Private Key used to sign the image
`container-image-signer-pgp-private-key` (deprecated)\
                       | Yes       |         | PGP Private Key used to sign the image
`container-image-tag`  | Yes       |         | Tag of the container image to sign
`container-image-signature-destination-url` \
                       | No        |         | URL to upload the container image signature to.\
                                               The container image signature name will be appended \
                                               to form the container image signature URI. \
                                               <br /><br />\
                                               Must start with `/` or `file://` to upload \
                                               to local file path. \
                                               <br /><br />\
                                               Or must start with `http://` or `https://` to \
                                               upload via a `PUT` to a remote location.
`container-image-signature-destination-username` \
                       | No        |         | Username to use when doing upload via http(s).
`container-image-signature-destination-password` \
                       | No        |         | Password to use when doing upload via http(s).
``

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key                        | Description
-------------------------------------------|------------
`container-image-signature-signer-private-key-fingerprint` \
                                           | Fingerprint for the private key used to sign \
                                             the container image.
`container-image-signed-tag`               | TODO
`container-image-signature-file-path`      | File path to created image signature.
`container-image-signature-name`           | Fully qualified name of the \
                                             name of the image signature, \
                                             including: organization, repo, \
                                             and hash. <br/>\
                                             ex: user/hello-node@sha256=\
                                             2cbdb73c9177e63e85d267f738e9\
                                             9e368db3f806eab4c541f5c6b719\
                                             e69f1a2b/signature-1
`container-image-signature-uri`            | URI of the uploaded container image signature.
`container-image-signature-upload-results` | Results of uploading the container image \
                                             signature to the given destination.

"""
import glob
import os
import sys
from distutils import util

import sh
from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.containers import container_registries_login
from ploigos_step_runner.utils.file import upload_file
from ploigos_step_runner.utils.pgp import import_pgp_key

DEFAULT_CONFIG = {
    'src-tls-verify': 'true'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    # signer-pgp-private-key - new key name
    # container-image-signer-pgp-private-key - old key name
    ['signer-pgp-private-key', 'container-image-signer-pgp-private-key'],

    # being flexible for different use cases of proceeding steps
    ['container-image-push-tag', 'container-image-tag']
]

class PodmanSign(StepImplementer):
    """`StepImplementer` for the `sign-container-image` step using Podman to create
    an image signature.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # get the pgp private key to sign the image with
        signer_pgp_private_key = self.get_value(
            ['signer-pgp-private-key', 'container-image-signer-pgp-private-key']
        )

        # get the uri to the image to sign
        container_image_tag = self.get_value(['container-image-push-tag', 'container-image-tag'])

        image_signatures_directory = self.create_working_dir_sub_dir(
            sub_dir_relative_path='image-signature'
        )
        try:
            # import the PGP key and get the finger print
            signer_pgp_private_key_fingerprint = import_pgp_key(
                pgp_private_key=signer_pgp_private_key
            )
            step_result.add_artifact(
                name='container-image-signature-signer-private-key-fingerprint',
                value=signer_pgp_private_key_fingerprint
            )

            # login to any provider container registries
            # NOTE 1: can not use auth file, even though we want to, because podman image sign
            #         does not accept authfile.
            #         https://github.com/containers/podman/issues/10866
            # NOTE 2: have to force login to use podman because even though logging in with
            #         any of the tools should work, in testing the podman sign only worked
            #         from within the python virtual environment if the login happened with podman.
            container_registries_login(
                registries=self.get_value('container-registries'),
                containers_config_tls_verify=util.strtobool(self.get_value('src-tls-verify')),
                container_command_short_name='podman'
            )

            # sign the image
            signature_file_path = PodmanSign.__sign_image(
                pgp_private_key_fingerprint=signer_pgp_private_key_fingerprint,
                image_signatures_directory=image_signatures_directory,
                container_image_tag=container_image_tag
            )
            step_result.add_artifact(
                name='container-image-signed-tag',
                value=container_image_tag,
            )
            step_result.add_artifact(
                name='container-image-signature-file-path',
                value=signature_file_path,
            )
            signature_name = os.path.relpath(signature_file_path, image_signatures_directory)
            step_result.add_artifact(
                name='container-image-signature-name',
                value=signature_name
            )

             # upload the image signature
            container_image_signature_destination_url = self.get_value(
                'container-image-signature-destination-url'
            )
            if container_image_signature_destination_url:
                container_image_signature_destination_uri = \
                    f"{container_image_signature_destination_url}/{signature_name}"
                step_result.add_artifact(
                    name='container-image-signature-uri',
                    description='URI of the uploaded container image signature',
                    value=container_image_signature_destination_uri
                )

                upload_result = upload_file(
                    file_path=signature_file_path,
                    destination_uri=container_image_signature_destination_uri,
                    username=self.get_value('container-image-signature-destination-username'),
                    password=self.get_value('container-image-signature-destination-password')
                )
                step_result.add_artifact(
                    name='container-image-signature-upload-results',
                    description='Results of uploading the container image signature' \
                                ' to the given destination.',
                    value=upload_result
                )
        except (RuntimeError, StepRunnerException) as error:
            step_result.success = False
            step_result.message = str(error)

        return step_result

    @staticmethod
    def __sign_image(
        pgp_private_key_fingerprint,
        image_signatures_directory,
        container_image_tag
    ):
        # sign image
        print(
            f"Sign image ({container_image_tag}) "
            f"with PGP private key ({pgp_private_key_fingerprint})"
        )
        try:
            # NOTE: for some reason the output from podman sign goes to stderr so....
            #       merge the two streams
            sh.podman.image( # pylint: disable=no-member
                "sign",
                f"--sign-by={pgp_private_key_fingerprint}",
                f"--directory={image_signatures_directory}",
                f"docker://{container_image_tag}",
                _out=sys.stdout,
                _err_to_out=True,
                _tee='out'
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error signing image ({container_image_tag}): {error}"
            ) from error

        # get image signature file path
        signature_file_paths = glob.glob(
            f"{image_signatures_directory}/**/signature-*",
            recursive=True
        )

        if len(signature_file_paths) != 1:
            raise StepRunnerException(
                f"Unexpected number of signature files, expected 1: {signature_file_paths}"
            )

        signature_file_path = signature_file_paths[0]
        print(
            f"Signed image ({container_image_tag}) with PGP private key "
            f"({pgp_private_key_fingerprint}): '{signature_file_path}'"
        )

        return signature_file_path



import unittest
import shutil

from ploigos_step_runner.decryption_utils import DecryptionUtils

class BaseTestCase(unittest.TestCase):
    def setUp(self):
        try:
            self.maxDiff = None
            shutil.rmtree("./step-runner-working")
        except FileNotFoundError:
            pass

    def tearDown(self):
        DecryptionUtils._DecryptionUtils__config_value_decryptors = []
        DecryptionUtils._DecryptionUtils__obfuscation_streams = []

        try:
            shutil.rmtree("./step-runner-working")
        except FileNotFoundError:
            pass


import os.path

from ploigos_step_runner.config import Config, ConfigValue
from ploigos_step_runner.config.decryptors.sops import SOPS
from ploigos_step_runner.decryption_utils import DecryptionUtils
from testfixtures import TempDirectory
from tests.helpers.base_test_case import BaseTestCase


class TestConfig(BaseTestCase):
    def test_add_config_invalid_type(self):
        config = Config()
        with self.assertRaisesRegex(
            ValueError,
            r"Given config \(True\) is unexpected type \(<class 'bool'>\) not a dictionary, string, or list of former."
        ):
            config.add_config(True)

    def test_add_config_dict_missing_config_key(self):
        config = Config()
        with self.assertRaisesRegex(
            AssertionError,
            r"Failed to add invalid config. Missing expected top level key \(step-runner-config\):"
        ):
            config.add_config({
                'foo': 'foo'
            })

    def test_add_config_dict_valid_basic(self):
        config = Config()
        config.add_config({
            Config.CONFIG_KEY: {}
        })

        self.assertEqual(config.global_defaults, {})
        self.assertEqual(config.global_environment_defaults, {})

    def test_initial_config_dict_valid_basic(self):
        config = Config({
            Config.CONFIG_KEY: {}
        })

        self.assertEqual(config.global_defaults, {})
        self.assertEqual(config.global_environment_defaults, {})

    def test_initial_config_dict_missing_config_key(self):
        with self.assertRaisesRegex(
            AssertionError,
            r"Failed to add invalid config. Missing expected top level key \(step-runner-config\):"
        ):
            Config({
                'foo': 'foo'
            })

    def test_add_config_file_missing_file(self):
        with TempDirectory() as temp_dir:
            config = Config()
            with self.assertRaisesRegex(
                ValueError,
                r"Given config string \(.*\) is not a valid path."
            ):
                config.add_config(os.path.join(temp_dir.path, 'does-not-exist.yml'))

    def test_add_config_file_invalid_json_or_yaml(self):
        with TempDirectory() as temp_dir:
            config_file_name = "bad"
            config_file_contents = ": blarg this: is {} bad syntax"
            temp_dir.write(config_file_name, bytes(f"{config_file_contents}", 'utf-8'))

            config = Config()
            with self.assertRaisesRegex(
                ValueError,
                r"Error parsing config file \(.*\) as json or yaml"
            ):
                config.add_config(os.path.join(temp_dir.path, config_file_name))

    def test_add_config_file_missing_config_key(self):
        with TempDirectory() as temp_dir:
            config_file_name = "foo.json"
            config_file_contents = {
                'foo': 'foo'
            }
            temp_dir.write(config_file_name, bytes(f"{config_file_contents}", 'utf-8'))

            config = Config()
            with self.assertRaisesRegex(
                AssertionError,
                r"Failed to add parsed configuration file \(.*\): Failed to add invalid config. Missing expected top level key \(step-runner-config\):"
            ):
                config.add_config(os.path.join(temp_dir.path, config_file_name))

    def test_add_config_file_valid_basic(self):
        with TempDirectory() as temp_dir:
            config_file_name = "foo.json"
            config_file_contents = {
                Config.CONFIG_KEY: {}
            }
            temp_dir.write(config_file_name, bytes(f"{config_file_contents}", 'utf-8'))

            config = Config()
            config.add_config(os.path.join(temp_dir.path, config_file_name))

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})

    def test_add_config_dir_no_files(self):
        with TempDirectory() as temp_dir:
            config = Config()
            with self.assertRaisesRegex(
                ValueError,
                r"Given config string \(.*\) is a directory with no recursive children files."
            ):
                config.add_config(temp_dir.path)

    def test_add_config_dir_one_valid_file(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {}
                    }
                }
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(os.path.join(temp_dir.path, config_dir))

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})

    def test_add_config_dir_two_valid_file(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : {
                                'implementer': 'foo'
                            }
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-bar' : {
                                'implementer': 'bar'
                            }
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(os.path.join(temp_dir.path, config_dir))

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})

    def test_add_two_valid_files(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : {
                                'implementer': 'foo',
                                'config': {
                                  'f': 'hello1'
                                }
                            }
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-bar' : {
                                'implementer': 'bar',
                                'config': {
                                  'b': 'hello2'
                                }
                            }
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(
                [
                    os.path.join(temp_dir.path, config_dir, 'foo.yml'),
                    os.path.join(temp_dir.path, config_dir, 'bar.yml')
                ]
            )

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('foo').get_config_value('f'),
                'hello1'
            )
            self.assertEqual(
                config.get_step_config('step-test-bar').get_sub_step('bar').get_config_value('b'),
                'hello2'
            )

    def test_add_two_valid_files_one_sub_step_diff_config(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : [
                                 {
                                    'implementer': 'foo',
                                    'config': {
                                        'foo': 'hello'
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : [
                                 {
                                    'implementer': 'foo',
                                    'config': {
                                        'bar': 'world'
                                    }
                                }
                            ]
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(
                [
                    os.path.join(temp_dir.path, config_dir, 'foo.yml'),
                    os.path.join(temp_dir.path, config_dir, 'bar.yml')
                ]
            )

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('foo').get_config_value('foo'),
                'hello'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('foo').get_config_value('bar'),
                'world'
            )

    def test_add_two_valid_files_two_sub_step_diff_config(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : [
                                 {
                                    'implementer': 'a',
                                    'config': {
                                        'a1': 'hello-a1'
                                    }
                                },
                                {
                                    'implementer': 'b',
                                    'config': {
                                        'b1': 'hello-b1'
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : [
                                 {
                                    'implementer': 'a',
                                    'config': {
                                        'a2': 'hello-a2'
                                    }
                                },
                                {
                                    'implementer': 'b',
                                    'config': {
                                       'b2': 'hello-b2'
                                    }
                                }
                            ]
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(
                [
                    os.path.join(temp_dir.path, config_dir, 'foo.yml'),
                    os.path.join(temp_dir.path, config_dir, 'bar.yml')
                ]
            )

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('a').get_config_value('a1'),
                'hello-a1'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('b').get_config_value('b1'),
                'hello-b1'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('a').get_config_value('a2'),
                'hello-a2'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('b').get_config_value('b2'),
                'hello-b2'
            )

    def test_add_two_valid_files_two_sub_step_with_human_names_diff_config(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : [
                                 {
                                    'name': 'A human',
                                    'implementer': 'a',
                                    'config': {
                                        'a1': 'hello-a1'
                                    }
                                },
                                {
                                    'name': 'B human',
                                    'implementer': 'b',
                                    'config': {
                                        'b1': 'hello-b1'
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-test-foo' : [
                                 {
                                    'name': 'A human',
                                    'implementer': 'a',
                                    'config': {
                                        'a2': 'hello-a2'
                                    }
                                },
                                {
                                    'name': 'B human',
                                    'implementer': 'b',
                                    'config': {
                                        'b2': 'hello-b2'
                                    }
                                }
                            ]
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(
                [
                    os.path.join(temp_dir.path, config_dir, 'foo.yml'),
                    os.path.join(temp_dir.path, config_dir, 'bar.yml')
                ]
            )

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('A human').get_config_value('a1'),
                'hello-a1'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('B human').get_config_value('b1'),
                'hello-b1'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('A human').get_config_value('a2'),
                'hello-a2'
            )
            self.assertEqual(
                config.get_step_config('step-test-foo').get_sub_step('B human').get_config_value('b2'),
                'hello-b2'
            )

    def test_add_config_dir_one_valid_file_one_invalid_file(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {}
                    }
                },
                {
                    'name': os.path.join(config_dir,'bad.yml'),
                    'contents' : {
                        'bad': {
                            'implementer': 'bar'
                        }
                    }
                }
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            with self.assertRaisesRegex(
                AssertionError,
                r"Failed to add parsed configuration file \(.*\): Failed to add invalid config. Missing expected top level key \(step-runner-config\):"
            ):
                config.add_config(os.path.join(temp_dir.path, config_dir))

    def test_add_config_dir_one_valid_file_and_one_valid_dict(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'step-foo': {
                                'implementer': 'foo'
                            }
                        }
                    }
                }
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(
                [
                    os.path.join(temp_dir.path, config_dir),
                    {
                        Config.CONFIG_KEY: {
                            'step-bar': {
                                'implementer': 'bar'
                            }
                        }
                    }
                ]

            )

            self.assertEqual(config.global_defaults, {})
            self.assertEqual(config.global_environment_defaults, {})

    def test_initial_config_dict_step_missing_implementer(self):
        with self.assertRaisesRegex(
            AssertionError,
            r"Step \(invalid-step\) defines a single sub step with values \({}\) but is missing value for key: implementer"
        ):
            Config({
                Config.CONFIG_KEY: {
                    'invalid-step' : {}
                }
            })

    def test_global_defaults(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-defaults' : {
                    'test1' : 'foo1'
                }
            }
        })

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(config.global_defaults),
            {
                'test1' : 'foo1'
            }
        )
        self.assertEqual(config.global_environment_defaults, {})

    def test_global_environment_defaults(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-environment-defaults' : {
                    'env1': {
                        'test1': 'env1',
                        'test2': 'test2'
                    },
                    'env2': {
                        'test1': 'env2',
                        'test3': 'test3'
                    }
                }
            }
        })

        self.assertEqual(config.global_defaults, {})
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(config.global_environment_defaults),
            {
                'env1': {
                    'environment-name' : 'env1',
                    'test1' : 'env1',
                    'test2' : 'test2'
                },
                'env2': {
                    'environment-name' : 'env2',
                    'test1' : 'env2',
                    'test3' : 'test3'
                }
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                config.get_global_environment_defaults_for_environment('env1')
            ),
            {
                'environment-name' : 'env1',
                'test1' : 'env1',
                'test2' : 'test2'
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                config.get_global_environment_defaults_for_environment('env2')
            ), {
                'environment-name' : 'env2',
                'test1' : 'env2',
                'test3' : 'test3'
            }
        )
        self.assertEqual(config.get_global_environment_defaults_for_environment('does-not-exist'), {})
        self.assertEqual(config.get_global_environment_defaults_for_environment(None), {})

    def test_get_sub_step_configs_for_non_existent_step(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': {
                    'implementer': 'foo'
                }
            }
        })

        self.assertEqual(config.get_sub_step_configs('does-not-exist'), [])

    def test_get_sub_step_configs_single_sub_step_no_config(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': {
                    'implementer': 'foo',
                }
            }
        })

        sub_step_configs = config.get_sub_step_configs('step-foo')
        self.assertEqual(len(sub_step_configs), 1)
        sub_step_config = sub_step_configs[0]
        self.assertEqual(sub_step_config.sub_step_config, {})

    def test_set_step_config_overrides_existing_step(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': {
                    'implementer': 'foo',
                }
            }
        })
        sub_step_configs = config.get_sub_step_configs('step-foo')
        self.assertEqual(len(sub_step_configs), 1)
        sub_step_config = sub_step_configs[0]
        self.assertEqual(sub_step_config.step_config_overrides, {})

        config.set_step_config_overrides('step-foo', {
            'test1': 'test2'
        })

        self.assertEqual(sub_step_config.step_config_overrides, {
            'test1': 'test2'
        })

    def test_set_step_config_override_overrides_existing_step(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': {
                    'implementer': 'foo',
                }
            }
        })
        sub_step_configs = config.get_sub_step_configs('step-foo')
        self.assertEqual(len(sub_step_configs), 1)
        sub_step_config = sub_step_configs[0]
        self.assertEqual(sub_step_config.step_config_overrides, {})

        config.set_step_config_overrides('step-foo', {
            'test1': 'test2'
        })
        self.assertEqual(sub_step_config.step_config_overrides, {
            'test1': 'test2'
        })

        config.set_step_config_overrides('step-foo', {
            'test2': 'test3'
        })
        self.assertEqual(sub_step_config.step_config_overrides, {
            'test2': 'test3'
        })

    def test_set_step_config_overrides_not_existing_step(self):
        config = Config({
            Config.CONFIG_KEY: {
            }
        })
        self.assertIsNone(config.get_step_config('step-bar'))

        config.set_step_config_overrides('step-bar', {
            'test1': 'test2'
        })

        step_config = config.get_step_config('step-bar')
        self.assertIsNotNone(step_config)
        self.assertEqual(step_config.step_config_overrides, {
            'test1': 'test2'
        })

    def test_duplicate_global_default_keys(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'global-defaults' : {
                                'dup-key': 'foo'
                            }
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'global-defaults' : {
                                'dup-key': 'bar'
                            }
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            with self.assertRaisesRegex(
                ValueError,
                r"Error merging global defaults: Conflict at dup-key"
            ):
                config = Config()
                config.add_config(os.path.join(temp_dir.path, config_dir))

    def test_duplicate_global_environment_default_keys(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'global-environment-defaults' : {
                                'env1' : {
                                    'dup-key': 'foo'
                                }
                            }
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'global-environment-defaults' : {
                                'env1' : {
                                    'dup-key': 'bar'
                                }
                            }
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            with self.assertRaisesRegex(
                ValueError,
                r"Error merging global environment \(env1\) defaults: Conflict at dup-key"
            ):
                config = Config()
                config.add_config(os.path.join(temp_dir.path, config_dir))

    def test_merge_valid_global_environment_defaults_same_env_diff_keys(self):
        with TempDirectory() as temp_dir:
            config_dir = "test"

            config_files = [
                {
                    'name': os.path.join(config_dir,'foo.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'global-environment-defaults' : {
                                'env1' : {
                                    'foo-key': 'foo'
                                }
                            }
                        }
                    }
                },
                {
                    'name': os.path.join(config_dir,'bar.yml'),
                    'contents' : {
                        Config.CONFIG_KEY: {
                            'global-environment-defaults' : {
                                'env1' : {
                                    'bar-key': 'bar'
                                }
                            }
                        }
                    }
                },
            ]

            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']
                temp_dir.write(
                    config_file_name,
                    bytes(f"{config_file_contents}", 'utf-8')
                )

            config = Config()
            config.add_config(os.path.join(temp_dir.path, config_dir))
            self.assertEqual(
                ConfigValue.convert_leaves_to_values(
                    config.get_global_environment_defaults_for_environment('env1')
                ),
                {
                    'environment-name': 'env1',
                    'foo-key': 'foo',
                    'bar-key': 'bar'
                }
            )
    def test_invalid_sub_steps(self):
        with self.assertRaisesRegex(
            ValueError,
            r"Expected step \(step-foo\) to have have step config " \
            r"\(ConfigValue\(value=bad-step-config, " \
            r"value_path='\['step-runner-config', 'step-foo'\]\'\)\) of " \
            r"type dict or list but got: <class 'ploigos_step_runner.config.config_value.ConfigValue'>"
        ):
            Config({
                Config.CONFIG_KEY: {
                    'step-foo': "bad-step-config"
                }
            })

    def test_multiple_sub_steps(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    },
                    {
                        'implementer': 'foo2',
                        'config': {
                            'test2': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 2)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').sub_step_config,
            ),
            {
                'test1': 'foo'
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo2').sub_step_config
            ),
            {
                'test2': 'foo'
            }
        )

    def test_sub_step_with_continue_sub_steps_on_failure_bool(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'continue-sub-steps-on-failure': True,
                        'config': {
                            'test1': 'foo'
                        }
                    },
                    {
                        'implementer': 'foo2',
                        'config': {
                            'test2': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 2)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').sub_step_config,
            ),
            {
                'test1': 'foo'
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo2').sub_step_config
            ),
            {
                'test2': 'foo'
            }
        )
        self.assertTrue(
            step_config.get_sub_step('foo1').sub_step_contine_sub_steps_on_failure
        )
        self.assertFalse(
            step_config.get_sub_step('foo2').sub_step_contine_sub_steps_on_failure
        )

    def test_sub_step_with_continue_sub_steps_on_failure_str(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'continue-sub-steps-on-failure': 'true',
                        'config': {
                            'test1': 'foo'
                        }
                    },
                    {
                        'implementer': 'foo2',
                        'config': {
                            'test2': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 2)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').sub_step_config,
            ),
            {
                'test1': 'foo'
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo2').sub_step_config
            ),
            {
                'test2': 'foo'
            }
        )
        self.assertTrue(
            step_config.get_sub_step('foo1').sub_step_contine_sub_steps_on_failure
        )
        self.assertFalse(
            step_config.get_sub_step('foo2').sub_step_contine_sub_steps_on_failure
        )

    def test_sub_step_with_name(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'name': 'sub-step-name-test',
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 1)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('sub-step-name-test').sub_step_config
            ),
            {
                'test1': 'foo'
            }
        )

    def test_sub_step_with_environment_config(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        },
                        'environment-config': {
                            'env1': {
                                'test2': 'bar'
                            }
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 1)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').sub_step_config
            ),
            {
                'test1': 'foo'
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').get_sub_step_env_config('env1')
            ),
            {
                'test2': 'bar'
            }
        )

    def test_sub_step_with_no_environment_config(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 1)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').sub_step_config
            ),
            {
                'test1': 'foo'
            }
        )
        self.assertEqual(step_config.get_sub_step('foo1').get_sub_step_env_config('env1'), {})

    def test_parse_and_register_decryptors_definitions_one_definition(self):
        Config.parse_and_register_decryptors_definitions(
            [
                {
                    'implementer': 'SOPS'
                }
            ]
        )

        sops_decryptor = DecryptionUtils._DecryptionUtils__config_value_decryptors[0]
        self.assertEqual(
            type(sops_decryptor),
            SOPS
        )

    def test_parse_and_register_decryptors_definitions_one_definition_with_config(self):
        Config.parse_and_register_decryptors_definitions(
            [
                {
                    'implementer': 'SOPS',
                    'config': {
                        'additional_sops_args': [
                            '--aws-profile=foo'
                        ]
                    }
                }
            ]
        )

        sops_decryptor = DecryptionUtils._DecryptionUtils__config_value_decryptors[0]
        self.assertEqual(
            sops_decryptor._SOPS__additional_sops_args,
            ['--aws-profile=foo']
        )

    def test_initial_config_dict_valid_with_decryptor_definition(self):
        config = Config({
            'step-runner-config': {
                'config-decryptors': [
                    {
                        'implementer': 'SOPS',
                        'config': {
                            'additional_sops_args': [
                                '--aws-profile=foo'
                            ]
                        }
                    }
                ]
            }
        })

        self.assertEqual(config.global_defaults, {})
        self.assertEqual(config.global_environment_defaults, {})
        sops_decryptor = DecryptionUtils._DecryptionUtils__config_value_decryptors[0]
        self.assertEqual(
            type(sops_decryptor),
            SOPS
        )
        self.assertEqual(
            sops_decryptor._SOPS__additional_sops_args,
            ['--aws-profile=foo']
        )



"""`StepImplementer` for the `opa` step using Open Policy Agent.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key       | Required? | Default | Description
-------------------     |-----------|---------|-----------
`workflow-policy-uri`   | Yes       |         | URI to policy used to audit attestations
`evidence-uri`          | Yes       |         | URI to evidence file to be audited

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`audit-results`     | Results of opa audit of workflow attestations
"""

import sys
from io import StringIO
import sh
from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.utils.file import download_source_to_destination
from ploigos_step_runner.utils.io import \
    create_sh_redirect_to_multiple_streams_fn_callback


DEFAULT_CONFIG = {}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'workflow-policy-uri',
    'evidence-uri'
]
class OpenPolicyAgent(StepImplementer):  # pylint: disable=too-few-public-methods
    """`StepImplementer` for the OPA class.
    """

    DEFAULT_WORKFLOW_POLICY_QUERY = "data.workflowResult.passAll"
    DEFAULT_WORKFLOW_POLICY_DATA_QUERY = "data"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):

        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.

        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def __audit_attestation(self, # pylint: disable=no-self-use
        workflow_attestation_file_path,
        workflow_policy_file_path,
        workflow_policy_query):
        """Method to run the opa shell command to evaluate a data file with a provided
           query file and given query.

        Parameters
        ----------
        workflow_attestation_file_path
            File path to the workflow attestation file
        workflow_policy_file_path
            File path to the workflow policy file
        workflow_policy_query
            Query that is run against the policy file

        Returns
        -------
        string
            Output of the opa shell command. This is more important
            to output if the query fails.
        int
            Return code for opa shell command. A value of 0
            means the query succeeeded whereas a value of
            1 means the query failed.
        """

        opa_attestation_stdout_result = StringIO()
        opa_attestation_stdout_callback = create_sh_redirect_to_multiple_streams_fn_callback([
            sys.stdout,
            opa_attestation_stdout_result
        ])

        try:
            sh.opa( # pylint: disable=no-member
                'eval',
                '--fail-defined',
                '-d',
                workflow_attestation_file_path,
                '-i',
                workflow_policy_file_path,
                workflow_policy_query,
                _out=opa_attestation_stdout_callback,
                _err_to_out=True,
                _tee='out'
            )

        except sh.ErrorReturnCode as error:
            return f"Error evaluating query against data:  {error}", 1

        return 'Audit was successful', 0

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        work_dir = self.work_dir_path

        #workflow attestation uri
        workflow_attestation_uri = self.get_value('evidence-uri')

        if workflow_attestation_uri is None:
            step_result.success = False
            step_result.message = 'No value found for evidence-uri'
            return step_result

        workflow_attestation_file_path = download_source_to_destination(
            workflow_attestation_uri,
            work_dir
        )

        workflow_policy_uri = self.get_value('workflow-policy-uri')

        #Download workflow policy from configured uri
        workflow_policy_file_path = download_source_to_destination(
            workflow_policy_uri,
            work_dir)

        audit_results, return_code = self.__audit_attestation(workflow_attestation_file_path,
            workflow_policy_file_path,
            self.DEFAULT_WORKFLOW_POLICY_QUERY)

        if return_code == 1:
            step_result.success = False
            step_result.message = "Attestation error: " + audit_results

            detailed_report, return_code = self.__audit_attestation(workflow_attestation_file_path,
                workflow_policy_file_path,
                self.DEFAULT_WORKFLOW_POLICY_DATA_QUERY)
            audit_results = detailed_report

        else:
            step_result.message = "Audit was successful"

        step_result.add_artifact(
                name='audit-results',
                value=audit_results
        )

        return step_result



import os
from io import IOBase
from pathlib import Path
from unittest.mock import patch

import mock
import re
import sh
from ploigos_step_runner import StepResult
from ploigos_step_runner.results.workflow_result import WorkflowResult
from ploigos_step_runner.step_implementers.audit_attestation import OpenPolicyAgent
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any, create_sh_side_effect

class TestStepImplementerOpenPolicyAgentBase(BaseStepImplementerTestCase):

    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path='',
            environment=None
    ):
        return self.create_given_step_implementer(
            step_implementer=OpenPolicyAgent,
            step_config=step_config,
            step_name='audit_attestation',
            implementer='OpenPolicyAgent',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            environment=environment
        )

class TestStepImplementerOpenPolicyAgent_other(TestStepImplementerOpenPolicyAgentBase):
    def test_step_implementer_config_defaults(self):
        defaults = OpenPolicyAgent.step_implementer_config_defaults()
        expected_defaults = {
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = OpenPolicyAgent._required_config_or_result_keys()
        expected_required_keys = [
            'workflow-policy-uri',
            'evidence-uri'
        ]
        self.assertEqual(required_keys, expected_required_keys)

class TestStepImplementerOpenPolicyAgent_run_step(TestStepImplementerOpenPolicyAgentBase):


    @patch('ploigos_step_runner.step_implementers.audit_attestation.opa.download_source_to_destination')
    @patch.object(
        OpenPolicyAgent,
        '_OpenPolicyAgent__audit_attestation',
        return_value = None
    )
    def test__run_step_pass_audit_success(self,
        audit_attestation_mock,
        download_source_to_destination_mock):
        with TempDirectory() as temp_dir:

            workflow_attestation_uri = 'https://foo.bar/evidence.json'
            workflow_policy_uri = 'https://foo.bar/policy.json'

            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'workflow-policy-uri': workflow_policy_uri
            }

            step_result = StepResult(
            step_name='test-step',
            sub_step_name='test-sub-step',
            sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact('evidence-uri', workflow_attestation_uri, 'URI of the uploaded results archive.')

            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            download_source_to_destination_mock.side_effect = [
                parent_work_dir_path + '/workflow_attestion.json',
                parent_work_dir_path + '/workflow_policy.rego']

            audit_attestation_mock.return_value = "Audit was successful", 0

            step_result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='audit_attestation',
                sub_step_name='OpenPolicyAgent',
                sub_step_implementer_name='OpenPolicyAgent'
            )
            expected_step_result.add_artifact(
                name='audit-results',
                value='Audit was successful'
            )
            expected_step_result.message = 'Audit was successful'

            self.assertEqual(step_result, expected_step_result)

            audit_attestation_mock.assert_called_once_with(parent_work_dir_path + '/workflow_attestion.json',
                parent_work_dir_path + '/workflow_policy.rego', "data.workflowResult.passAll")

            download_source_to_destination_mock.assert_has_calls([
                mock.call(workflow_attestation_uri,
                parent_work_dir_path + '/audit_attestation'),
                mock.call(workflow_policy_uri,
                parent_work_dir_path  + '/audit_attestation')
                ]
            )


    def test__run_step_fail_audit_fail_missing_workflow_attestation(self):
        with TempDirectory() as temp_dir:

            workflow_attestation_uri = 'https://foo.bar/evidence.json'
            workflow_policy_uri = 'https://foo.bar/policy.json'

            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'workflow-policy-uri': workflow_policy_uri
            }

            step_result = StepResult(
            step_name='test-step',
            sub_step_name='test-sub-step',
            sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact('evidence-uri-wrong-key', workflow_attestation_uri, 'URI of the uploaded results archive.')

            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            step_result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='audit_attestation',
                sub_step_name='OpenPolicyAgent',
                sub_step_implementer_name='OpenPolicyAgent'
            )

            expected_step_result.success = False
            expected_step_result.message = "No value found for evidence-uri"

            self.assertEqual(step_result, expected_step_result)

    @patch('ploigos_step_runner.step_implementers.audit_attestation.opa.download_source_to_destination')
    @patch.object(
        OpenPolicyAgent,
        '_OpenPolicyAgent__audit_attestation',
        return_value = None
    )
    def test__run_step_fail_audit_fail_error_on_audit(self,
        audit_attestation_mock,
        download_source_to_destination_mock):
        with TempDirectory() as temp_dir:

            workflow_attestation_uri = 'https://foo.bar/evidence.json'
            workflow_policy_uri = 'https://foo.bar/policy.json'

            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'workflow-policy-uri': workflow_policy_uri
            }

            step_result = StepResult(
            step_name='test-step',
            sub_step_name='test-sub-step',
            sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact('evidence-uri', workflow_attestation_uri, 'URI of the uploaded results archive.')

            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            download_source_to_destination_mock.side_effect = [
                parent_work_dir_path + '/workflow_attestion.json',
                parent_work_dir_path + '/workflow_policy.rego']

            audit_attestation_mock.return_value = "foo", 1

            step_result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='audit_attestation',
                sub_step_name='OpenPolicyAgent',
                sub_step_implementer_name='OpenPolicyAgent'
            )
            expected_step_result.add_artifact(
                name='audit-results',
                value='foo'
            )
            expected_step_result.success = False
            expected_step_result.message = step_result.message = "Attestation error: foo"

            self.assertEqual(step_result, expected_step_result)

            audit_attestation_mock.assert_has_calls([
                mock.call(parent_work_dir_path + '/workflow_attestion.json',
                parent_work_dir_path + '/workflow_policy.rego', "data.workflowResult.passAll"),
                mock.call(parent_work_dir_path + '/workflow_attestion.json',
                parent_work_dir_path + '/workflow_policy.rego', "data")
            ])

            download_source_to_destination_mock.assert_has_calls([
                mock.call(workflow_attestation_uri,
                parent_work_dir_path + '/audit_attestation'),
                mock.call(workflow_policy_uri,
                parent_work_dir_path  + '/audit_attestation')
                ]
            )

class TestStepImplementerOpenPolicyAgent_audit_attestation(TestStepImplementerOpenPolicyAgentBase):

    @patch('sh.opa', create=True)
    def test_audit_attestation_pass(self, audit_results_mock):


        step_config = {
                'workflow-policy-uri': "https://foo.bar"
        }

        step_implementer = self.create_step_implementer(
            step_config=step_config
        )
        audit_results_mock.return_value = "pass"

        workflow_attestation_file_path = "workflow_attestation_file_path"
        workflow_policy_file_path = "workflow_policy_file_path"
        workflow_policy_query =  "fooquery"

        step_implementer._OpenPolicyAgent__audit_attestation(
            workflow_attestation_file_path = "workflow_attestation_file_path",
            workflow_policy_file_path = "workflow_policy_file_path",
            workflow_policy_query =  "fooquery"
        )

        audit_results_mock.assert_called_once_with(
            'eval',
            '--fail-defined',
            '-d',
            workflow_attestation_file_path,
            '-i',
            workflow_policy_file_path,
            workflow_policy_query,
            _out=Any(IOBase),
            _err_to_out=True,
            _tee='out'
        )

    @patch('sh.opa', create=True)
    def test_audit_attestation_fail(self, audit_results_mock):


        step_config = {
                'workflow-policy-uri': "https://foo.bar"
        }

        step_implementer = self.create_step_implementer(
            step_config=step_config
        )
        audit_results_mock.return_value = "pass"
        audit_results_mock.side_effect = sh.ErrorReturnCode('opa', b'mock out', b'mock error')

        result_message, result_return_code = step_implementer._OpenPolicyAgent__audit_attestation(
            workflow_attestation_file_path = "workflow_attestation_file_path",
            workflow_policy_file_path = "workflow_policy_file_path",
            workflow_policy_query =  "fooquery"
        )

        self.assertRegex(result_message, re.compile(
                r"Error evaluating query against data:"
                r".*RAN: opa"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock error",
                re.DOTALL
            ))

        self.assertEqual(result_return_code, 1)



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os

from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from ploigos_step_runner.step_implementers.generate_metadata import Maven
from ploigos_step_runner import StepResult


class TestStepImplementerMavenGenerateMetadata(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Maven,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_step_implementer_config_defaults(self):
        defaults = Maven.step_implementer_config_defaults()
        expected_defaults = {
            'pom-file': 'pom.xml'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = Maven._required_config_or_result_keys()
        expected_required_keys = ['pom-file']
        self.assertEqual(required_keys, expected_required_keys)

    def test__validate_required_config_or_previous_step_result_artifact_keys_valid(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            temp_dir.write('pom.xml', b'''<project>
                <modelVersion>4.0.0</modelVersion>
                <groupId>com.mycompany.app</groupId>
                <artifactId>my-app</artifactId>
                <version>42.1</version>
            </project>''')
            pom_file_path = os.path.join(temp_dir.path, 'pom.xml')

            step_config = {
                'pom-file': pom_file_path
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Maven',
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_package_file_does_not_exist(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            pom_file_path = os.path.join(temp_dir.path, 'pom.xml')

            step_config = {
                'pom-file': pom_file_path
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Maven',
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                AssertionError,
                rf"Given pom file \(pom-file\) does not exist: {pom_file_path}"
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_run_step_pass(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            temp_dir.write('pom.xml', b'''<project>
                <modelVersion>4.0.0</modelVersion>
                <groupId>com.mycompany.app</groupId>
                <artifactId>my-app</artifactId>
                <version>42.1</version>
            </project>''')
            pom_file_path = os.path.join(temp_dir.path, 'pom.xml')

            step_config = {
                'pom-file': pom_file_path
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Maven',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Maven',
                sub_step_implementer_name='Maven'
            )
            expected_step_result.add_artifact(name='app-version', value='42.1')

            self.assertEqual(result, expected_step_result)

    def test_run_step_fail_missing_version_in_pom_file(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            temp_dir.write('pom.xml', b'''<project>
                <modelVersion>4.0.0</modelVersion>
                <groupId>com.mycompany.app</groupId>
                <artifactId>my-app</artifactId>
            </project>''')
            pom_file_path = os.path.join(temp_dir.path, 'pom.xml')

            step_config = {
                'pom-file': pom_file_path
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Maven',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Maven',
                sub_step_implementer_name='Maven'
            )
            expected_step_result.success = False
            expected_step_result.message = f"Given pom file ({pom_file_path}) does not contain " + \
                "a \"version\" key."

            self.assertEqual(result, expected_step_result)



# pylint: disable=line-too-long
"""Ploigos Step Runner (psr) main entry point.

Command-Line Options
--------------------

    -h, --help
        show this help message and exit

    -s STEP, --step STEP
        Ploigos workflow step to run

    -e ENVIRONMENT, --environment  ENVIRONMENT
        The environment to run this step against.

    -c CONFIG [CONFIG ...], --config CONFIG [CONFIG ...]
        Ploigos workflow configuration files, or directories containing files, in yml or json

    --step-config STEP_CONFIG_KEY=STEP_CONFIG_VALUE [STEP_CONFIG_KEY=STEP_CONFIG_VALUE ...]
        Override step config provided by the given Ploigos
        config-file with these arguments.

Step Configuration
------------------

### Steps

* generate-metadata
* tag-source
* static-code-analysis
* package
* unit-test
* push-artifacts
* create-container-image
* push-container-image
* sign-container-image
* container-image-unit-test
* container-image-static-compliance-scan
* container-image-static-vulnerability-scan
* create-deployment-environment
* deploy
* validate-environment-configuration
* uat
* runtime-vulnerability-scan
* canary-test
* report
* publish-workflow-results
* automated-governance

### Variable Precedence

From least precedence to highest precedence.

    1. StepImplementer implementation provided configuration defaults
    2. Global Configuration Defaults (step-runner-config.global-defaults)
    3. Global Environment Configuration Defaults (step-runner-config.global-environment-defaults)
    4. Step Configuration (step-runner-config.{STEP_NAME}.config)
    5. Step Environment Configuration
         (step-runner-config.{STEP_NAME}.environment-config.{ENVIRONMENT_NAME})
    6. Step Configuration Runtime Overrides (--environment arugment to main entry point)

** Example 1 **

    ---
    step-runner-config:
      # List of decryptors to use to decrypt any encrypted configuration.
      config-decryptors:
      - implementer: SOPS
        #config:
        #  additional_sops_args: [
        #    '--any-valid-sops-cmd-arg-here=value',
        #    '--aws-profile=FOO'
        #  ]

      # Dictionary of configuration options which will be used in step configuration if that
      # step does not have a specific value for that configuration already or one is not
      # given by global-environment-defaults.
      global-defaults:
        # A sample config option with a default global value.
        # Overrides
        #   * StepImplementers default config values
        # Is Overriden by:
        #   * global-environment-defaults
        #   * step config
        #   * step environment config
        #   * step config runtime overrides
        sample-config-option-1: 'global default'

      # Dictionary of dictionaries where the first level keys are environment names and their
      # dictionary values are configuration defaults to use when invoking a step in the context
      # of that environment.
      #
      # NOTE: Environment names can be anything so long as they line up with the environment value
      # given to the `--environment` flag of the main entry point.
      #global-environment-defaults:
        # Sample configuration for an environment named 'SAMPLE-ENV-1'.
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        #SAMPLE-ENV-1:
          # Sample config option that may differ from environment to environment.
          #sample-config-option-2: 'default for use in the SAMPLE-ENV-1 env"

        # Sample configuration for an environment named 'SAMPLE-ENV-2'.
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        #SAMPLE-ENV-2:
          # Sample config option that may differ from environment to environment.
          #sample-config-option-2: 'default for use in the SAMPLE-ENV-2 env"

      # Sample step config for step named SAMPLE-STEP-1
      SAMPLE-STEP-1:
      - implementer: SampleStep1Implementer1
        config:
          sample-config-option-3: 'value for sample-config-option-3 for use in this step'
          additional-artifacts-dirs:
          - name: sample-user-supplied-artifact
            path: sample/path/to/user/supplied/artifact
        environment-config:
          SAMPLE-ENV-1:
            sample-config-option-4: 'value for use in this step in SAMPLE-ENV-1 environment'
          SAMPLE-ENV-2:
            sample-config-option-4: 'value for use in this step in SAMPLE-ENV-1 environment'

### Example Configuration Files

.. Note::
    Optional step configurations are listed commented out and with their default values.

** Example Config file for a Maven built Application **

    ---
    step-runner-config:
      # Optional
      # List of decryptors to use to decrypt any encrypted configuration.
      config-decryptors:
      - implementer: SOPS
        #config:
        #  additional_sops_args: [
        #    '--any-valid-sops-cmd-arg-here=value',
        #    '--aws-profile=FOO'
        #  ]

      # Optional
      # Dictionary of configuration options which will be used in step configuration if that
      # step does not have a specific value for that configuration already or one is not
      # given by global-environment-defaults.
      global-defaults:
        # Required.
        # Name of the application the artifact built and deployed by this workflow is part of.
        application-name: ''

        # Required.
        # Name of the service this artifact built and deployed by this workflow implements as
        # part of the application it is a part of.
        service-name: ''

        # Optional.
        # Maven server settings for settings.xml file
        #maven-servers:
        #  internal-mirror-1:
        #    id: ''
        #    username: ''
        #    password: ''
        #  internal-mirror-2:
        #    id: ''
        #    username: ''
        #    password: ''

        # Optional.
        # Maven repository settings for settings.xml file
        #maven-repositories:
        #  internal-mirror-1:
        #    id: ''
        #    url: ''
        #    snapshots: ''
        #    releases: ''
        #  internal-mirror-2:
        #    id: ''
        #    url: ''
        #    snapshots: ''
        #    releases: ''

        # Optional.
        # Maven mirror settings for settings.xml file
        #maven-mirrors:
        #  internal-mirror-1:
        #    id: ''
        #    url: ''
        #    mirror-of: ''
        #  internal-mirror-2:
        #    id: ''
        #    url: ''
        #    mirror-of: ''

        # Dictionary of container registries to authenticate with.
        # Suggest putting in global configuration so it can be used for creating and pushing
        # images. But can also or instead be put in the individual steps if say different
        # registires are used for building images then pushing them.
        #container-registries:
        #  registry.redhat.io:
        #    username: account_number|acount_name
        #    password: encrypt_me
        #  registry.internal.example.xyz:
        #    username: team_name+project_name
        #    password: encrypt_me

      # Optional
      # Dictionary of dictionaries where the first level keys are environment names and their
      # dictionary values are configuration defaults to use when invoking a step in the context
      # of that environment.
      #
      # NOTE: Environment names can be anything so long as they line up with the environment value
      # given to the `--environment` flag of the main entry point.
      global-environment-defaults:
        # Sample
        # Sample configuration for an environment named 'DEV'.
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        DEV:
          #Optional
          #kube-api-token: ''

          # Required
          argocd-username: ''

          # Required
          argocd-password: ''

          # Required
          argocd-api: ''

          # Optional
          #argocd-sync-timeout-seconds: '60'

          # Optional
          #deployment-config-helm-chart-path: './'

        # Sample
        # Sample configuration for an environment named 'TEST'
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        TEST:
          #Optional
          #kube-api-token: ''

          # Required
          argocd-username: ''

          # Required
          argocd-password: ''

          # Required
          argocd-api: ''

          # Optional
          #argocd-sync-timeout-seconds: '60'

          # Optional
          #deployment-config-helm-chart-path: './'

        # Sample
        # Sample configuration for an environment named 'PROD'
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        #PROD:
        # Sample
        # Sample parameter that may differ from environment to environment.
        #kube-api-uri: 'api.prod.myorg.xyz"

      generate-metadata:
      - implementer: Maven
        config: {
          # Optional.
          #pom-file: 'pom.xml'
        }

      - implementer: Git
        config: {
          # Optional.
          #repo-root: './'

          # Optional.
          #build-string-length: 7
        }

      - implementer: SemanticVersion

      tag-source:
      - implementer: Git
        config: {
          # Optional.
          # Will use current directory to determine URL if not specified.
          #url: None

          # Optional.
          #username: None

          # Optional.
          #password: None
        }

      static-code-analysis:
      - implementer: SonarQube
        config: {
          # Required.
          # URL to the sonarqube server
          url: ''

          # Optional.
          # Properties file in root folder (eg: sonar-project.properties)
          #properties: ''

          # Optional.
          #user: None

          # Optional.
          #password: None
        }

      unit-test:
      - implementer: MavenTest
        config: {
          # Optional.
          # pom_file: 'pom.xml'

          # Optional.
          # tls-verify: True

          # Optional.
          # maven-profiles: []

          # Optional.
          # maven-no-transfer-progress: True

          # Optional.
          # maven-additional-arguments: []

          # Optional.
          # NOTE: Will attempt to dynamically determine from pom if not given.
          # test-reports-dir: 'target/surefire-reports
        }

      package:
      - implementer: MavenPackage
        config: {
          # Optional.
          #pom-file: 'pom.xml'

          # Optional
          #artifact-extensions: ['jar', 'war', 'ear']

          # Optional
          #artifact-parent-dir: 'target'
        }

      push-artifacts:
      - implementer: MavenDeploy
        config: {
          # Required.
          # URL to the artifact repository to push the artifact to.
          maven-push-artifact-repo-url: ''

          # Required.
          # Id to the artifact repository to push the artifact to.
          maven-push-artifact-repo-id: ''

        }

      create-container-image:
      - implementer: Buildah
        config: {
          # Optional.
          #imagespecfile: 'Dockerfile'

          # Optional.
          #context: '.'

          # Optional.
          #tls-verify: true

          # Optional.
          #format: 'oci'
        }

      generate-evidence:
      - implementer: GenerateEvidence
        config: {
          evidence-destination-url: 'http:/example-evidence-destination-url',
          evidence-destination-username: 'username'
        }
      - implementer: RekorSignEvidence
        config: {
          rekor-server-url: 'http://example-rekor-server-url-url/',
          signer-pgp-private-key: '/path/to/signer-pgp-private-key-path'
        }

      audit-attestation:
      - implementer: OpenPolicyAgent
        config: {
          workflow-policy-uri: 'http://example-workflow-policy-uri/uri',
          evidence-uri: 'http://example-evidence-uri/uri'
        }

      push-container-image:
      - implementer: Skopeo
        config: {
          destination: '' # Required. Container image repository destination to push image to
          #src-tls-verify: true # Optional
          #dest-tls-verify: true # Optional
        }

      sign-container-image:
      # sample signing container image but not pushing the signature anywhere
      #   signature will be included in workflow results archive.
      - implementer: PodmanSign
        config: {}

      # sample signing container image and pushing container image signature to remote repository
      #- implementer: PodmanSign
      #  config:
      #    container-image-signature-destination-url: https://repository.ploigos.com/container-image-signatures
      #    container-image-signature-destination-username: test-user
      #    container-image-signature-destination-password: pass123 # this value should be put in an encrypted config file

      # sample signing container image and pushing container image signature to local file
      #- implementer: PodmanSign
      #  config:
      #    container-image-signature-destination-url: file://image-signatures-mount/

      container-image-static-compliance-scan:
      # sample scans using DataStream file (preferred)
      - name: OpenSCAP - Compliance - Example Compliance Profile for UBI8 Container Images [Based on xccdf_org.ssgproject.content_profile_standard]
        implementer: OpenSCAP
        config:
          oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-ds.xml
          oscap-tailoring-uri: https://raw.githubusercontent.com/ploigos/ploigos-example-oscap-content/main/xccdf_com.redhat.ploigos_profile_example_ubi8-tailoring-xccdf.xml
          oscap-profile: xccdf_com.redhat.ploigos_profile_example_ubi8
      #- name: OpenSCAP - Compliance (Protection Profile for General Purpose Operating Systems) - DataStream
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-ds.xml
      #    oscap-profile: xccdf_org.ssgproject.content_profile_ospp
      #- name: OpenSCAP - Compliance (DISA STIG for Red hat Enterprise Linux 8) - DataStream
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-ds.xml
      #    oscap-profile: xccdf_org.ssgproject.content_profile_stig

      # sample scans using XCCDF file (okay if you don't have access to DataStream)
      #- name: OpenSCAP - Compliance (DISA STIG for Red hat Enterprise Linux 8) - XCCDF
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-xccdf.xml
      #    oscap-profile: stig
      #- name: OpenSCAP - Compliance (Protection Profile for General Purpose Operating Systems) - XCCDF
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-xccdf.xml
      #    oscap-profile: ospp

      container-image-static-vulnerability-scan:
      # sample vulnerability scan using XCCDF file (preferred)
      - name: OpenSCAP - Vulnerability - DataStream
        implementer: OpenSCAP
        config:
          oscap-input-definitions-uri: https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2

      # sample scans using OVAL file (if you must, but XCCDF is muuuuch better if availabe)
      #- name: OpenSCAP - Vulnerability - OVAL
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://www.redhat.com/security/data/oval/v2/RHEL8/rhel-8.oval.xml.bz2
      # sample scan using DataStream file

      deploy:
      - implementer: ArgoCD
        config:
          # argocd specific variables are set per environment above

          # Required
          helm-config-repo: ''

          # Optional
          #values-yaml-directory: './cicd/Deployment/'

          # Optional
          #value-yaml-template: 'values.yaml.j2'

          # Required
          git-email: ''

          # Optional
          #git-name: 'ploigos'

          # Optional
          #git-username: None

          # Optional
          #git-password: None

          # Any template parameters required by values.yaml.j2 can be listed below. Note dashes will
          # be converted to underscores to be compliant with the jinja template variable
          # specification
          readiness-probe-path: ''

      validate-environment-configuration:
      - implementer: ConfiglintFromArgocd
        config: {}
      - implementer: Configlint
        config: {
          # Optional.
          # Path to the rules file
          #rules: ''
        }

      uat:
      - implementer: MavenIntegrationTest
        config:
          target-host-url-maven-argument-name: 'target.base.url'
          maven-additional-arguments:
          - -Dselenium.hub.url=http://selenium.plogios.xyz:4242

          # Optional.
          # pom_file: 'pom.xml'

          # Optional.
          # tls-verify: True

          # Optional.
          # maven-profiles: []

          # Optional.
          # maven-no-transfer-progress: True

          # Optional.
          # maven-additional-arguments: []

          # Optional.
          # NOTE: Will attempt to dynamically determine from pom if not given.
          # test-reports-dir: 'target/surefire-reports

      report:
      - implementer: ResultArtifactsArchive
        config:
          results-archive-destination-url: https://artifact-repo.plogios.com/release-engineering-workflow-result-artifacts-archives/
          results-archive-destination-username: mock-name

          # NOTE: should encrypt this
          results-archive-destination-password: mock-pass

** Example Config file for a NPM built Application **

    ---
    step-runner-config:
      # Optional
      # List of decryptors to use to decrypt any encrypted configuration.
      config-decryptors:
      - implementer: SOPS
        #config:
        #  additional_sops_args: [
        #    '--any-valid-sops-cmd-arg-here=value',
        #    '--aws-profile=FOO'
        #  ]

      # Optional
      # Dictionary of configuration options which will be used in step configuration if that
      # step does not have a specific value for that configuration already or one is not
      # given by global-environment-defaults.
      global-defaults:
        # Required.
        # Name of the application the artifact built and deployed by this workflow is part of.
        application-name: ''

        # Required.
        # Name of the service this artifact built and deployed by this workflow implements as
        # part of the application it is a part of.
        service-name: ''

        # Optional.
        # Dictionary of container registries to authenticate with.
        # Suggest putting in global configuration so it can be used for creating and pushing
        # images. But can also or instead be put in the individual steps if say different
        # registires are used for building images then pushing them.
        #container-registries:
        #  registry.redhat.io:
        #    username: account_number|acount_name
        #    password: encrypt_me
        #  registry.internal.example.xyz:
        #    username: team_name+project_name
        #    password: encrypt_me

      # Optional
      # Dictionary of dictionaries where the first level keys are environment names and their
      # dictionary values are configuration defaults to use when invoking a step in the context
      # of that environment.
      #
      # NOTE: Environment names can be anything so long as they line up with the environment value
      # given to the `--environment` flag of the main entry point.
      global-environment-defaults:
        # Optional Sample
        # Sample configuration for an environment named 'DEV'.
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        DEV:
          #Optional
          #kube-api-token: ''

          # Required
          argocd-username: ''

          # Required
          argocd-password: ''

          # Required
          argocd-api: ''

          # Optional
          #argocd-sync-timeout-seconds: '60'

          # Optional
          #deployment-config-helm-chart-path: './'

        # Sample
        # Sample configuration for an environment named 'TEST'
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        TEST:
          #Optional
          #kube-api-token:

          # Required
          argocd-username: ''

          # Required
          argocd-password: ''

          # Required
          argocd-api: ''

          # Optional
          #argocd-sync-timeout-seconds: '60'

          # Optional
          #deployment-config-helm-chart-path: './'

        # Sample
        # Sample configuration for an environment named 'PROD'
        #
        # NOTE: Environment names can be anything so long as they line up with the environment value
        # given to the `--environment` flag of the main entry point.
        #PROD:
        # Sample
        # Sample parameter that may differ from environment to environment.
        #kube-api-uri: 'api.prod.myorg.xyz"

      generate-metadata:
      # WARNING: not yet implemented
      - implementer: NPM
        config: {}

      - implementer: Git
        config: {
          # Optional.
          #repo-root: './'

          # Optional.
          #build-string-length: 7
        }

      - implementer: SemanticVersion

      tag-source:
      - implementer: Git
        config: {
          # Optional.
          # Will use current directory to determine URL if not specified.
          #url: None

          # Optional.
          #username: None

          # Optional.
          #password: None
        }

      static-code-analysis:
      - implementer: SonarQube
        config: {
          # Required.
          # URL to the sonarqube server
          url: ''

          # Optional.
          # Properties file in root folder (eg: sonar-project.properties)
          #properties: ''

          # Optional.
          #user: None

          # Optional.
          #password: None
        }

      package:
      # WARNING: not yet implemented
      - implementer: NPM
        config: {}

      unit-test:
      # WARNING: not yet implemented
      - implementer: NPM
        config: {}

      push-artifacts:
      # WARNING: not yet implemented
      - implementer: NPM
        config: {}

      create-container-image:
      - implementer: Buildah
        config: {
          # Optional.
          #imagespecfile: 'Dockerfile'

          # Optional.
          #context: '.'

          # Optional.
          #tls-verify: true

          # Optional.
          #format: 'oci'
        }

      generate-evidence:
      - implementer: GenerateEvidence
        config: {
          evidence-destination-url: 'http:/example-evidence-destination-url',
          evidence-destination-username: 'username'
        }
      - implementer: RekorSignEvidence
        config: {
          rekor-server-url: 'http://example-rekor-server-url-url/',
          signer-pgp-private-key: '/path/to/signer-pgp-private-key-path'
        }

      audit-attestation:
      - implementer: OpenPolicyAgent
        config: {
          workflow-policy-uri: 'http://example-workflow-policy-uri/uri',
          evidence-uri: 'http://example-evidence-uri/uri'
        }

      push-container-image:
      - implementer: Skopeo
        config: {
          destination: '' # Required. Container image repository destination to push image to
          #src-tls-verify: true # Optional
          #dest-tls-verify: true # Optional
        }

      sign-container-image:
      # sample signing container image but not pushing the signature anywhere
      #   signature will be included in workflow results archive.
      - implementer: PodmanSign
        config: {}

      # sample signing container image and pushing container image signature to remote repository
      #- implementer: PodmanSign
      #  config:
      #    container-image-signature-destination-url: https://repository.ploigos.com/container-image-signatures
      #    container-image-signature-destination-username: test-user
      #    container-image-signature-destination-password: pass123 # this value should be put in an encrypted config file

      # sample signing container image and pushing container image signature to local file
      #- implementer: PodmanSign
      #  config:
      #    container-image-signature-destination-url: file://image-signatures-mount/

      container-image-static-compliance-scan:
      # sample scans using DataStream file (preferred)
      - name: OpenSCAP - Compliance - Example Compliance Profile for UBI8 Container Images [Based on xccdf_org.ssgproject.content_profile_standard]
        implementer: OpenSCAP
        config:
          oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-ds.xml
          oscap-tailoring-uri: https://raw.githubusercontent.com/ploigos/ploigos-example-oscap-content/main/xccdf_com.redhat.ploigos_profile_example_ubi8-tailoring-xccdf.xml
          oscap-profile: xccdf_com.redhat.ploigos_profile_example_ubi8
      #- name: OpenSCAP - Compliance (Protection Profile for General Purpose Operating Systems) - DataStream
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-ds.xml
      #    oscap-profile: xccdf_org.ssgproject.content_profile_ospp
      #- name: OpenSCAP - Compliance (DISA STIG for Red hat Enterprise Linux 8) - DataStream
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-ds.xml
      #    oscap-profile: xccdf_org.ssgproject.content_profile_stig

      # sample scans using XCCDF file (okay if you don't have access to DataStream)
      #- name: OpenSCAP - Compliance (DISA STIG for Red hat Enterprise Linux 8) - XCCDF
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-xccdf.xml
      #    oscap-profile: stig
      #- name: OpenSCAP - Compliance (Protection Profile for General Purpose Operating Systems) - XCCDF
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://atopathways.redhatgov.io/compliance-as-code/scap/ssg-rhel8-xccdf.xml
      #    oscap-profile: ospp

      container-image-static-vulnerability-scan:
      # sample vulnerability scan using XCCDF file (preferred)
      - name: OpenSCAP - Vulnerability - DataStream
        implementer: OpenSCAP
        config:
          oscap-input-definitions-uri: https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2

      # sample scans using OVAL file (if you must, but XCCDF is muuuuch better if availabe)
      #- name: OpenSCAP - Vulnerability - OVAL
      #  implementer: OpenSCAP
      #  config:
      #    oscap-input-definitions-uri: https://www.redhat.com/security/data/oval/v2/RHEL8/rhel-8.oval.xml.bz2
      # sample scan using DataStream file

      deploy:
      - implementer: ArgoCD
        config:
          # argocd specific variables are set per environment above

          # Required
          helm-config-repo: ''

          # Optional
          #values-yaml-directory: './cicd/Deployment/'

          # Optional
          #value-yaml-template: 'values.yaml.j2'

          # Optional
          #deployment-config-helm-chart-path: './'

          # Required
          git-email: ''

          # Optional
          #git-name: 'ploigos'

          # Optional
          #git-username: None

          # Optional
          #git-password: None

          # Any template parameters required by values.yaml.j2 can be listed below.
          # Note dashes will be converted to underscores to be compliant with the jinja
          # template variable specification
          readiness-probe-path: ''

      # TODO: not yet implemented
      uat: []

      report:
      - implementer: ResultArtifactsArchive
        config:
          results-archive-destination-url: https://artifact-repo.plogios.com/release-engineering-workflow-result-artifacts-archives/
          results-archive-destination-username: mock-name

          # NOTE: should encrypt this
          results-archive-destination-password: mock-pass

Examples
--------

Getting Help

>>> psr --help


Example Running the 'generate-metadata' step

>>> psr
...     --config=my-app-step-runner-config.yml \
...     --step=generate-metadata

"""

import __main__
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.results import StepResult, WorkflowResult
from ploigos_step_runner.step_implementer import DefaultSteps, StepImplementer
from ploigos_step_runner.step_runner import StepRunner



"""Shared utils for maven operations.
"""
import os
import sys
import xml.etree.ElementTree as ET

import sh
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.io import \
    create_sh_redirect_to_multiple_streams_fn_callback
from ploigos_step_runner.utils.xml import (get_xml_element_by_path,
                                           get_xml_element_text_by_path)


def generate_maven_settings(working_dir, maven_servers, maven_repositories, maven_mirrors):
    """
    Generates and returns a settings.xml file from the inputs it is provided.

    Parameters
    ----------
    maven_servers:
        Dictionary of id, username, password
    maven_repositories:
        Dictionary of id, url, snapshots, releases
    maven_mirrors:
        Dictionary of id, url, mirror_of

    Raises
    ------
    ValueError
        If the given file can not be parsed as YAML or JSON.

    Returns
    -------
    Str
        The path to the settings.xml that was generated.
    """

    root = ET.Element('settings')
    add_maven_servers(root, maven_servers)
    add_maven_repositories(root, maven_repositories)
    add_maven_mirrors(root, maven_mirrors)

    tree = ET.ElementTree(root)

    settings_path = working_dir + '/settings.xml'
    with open(settings_path, 'wb') as file:
        tree.write(file)

    return settings_path

def add_maven_servers(parent_element, maven_servers):
    """Adds <servers> element to parent element and creates <server> elements for each given
    maven server.

    Parameters
    ----------
    parent_element : ET.Element
        Parent element to add the <servers> element too.

    maven_servers : (dict, list)
        List of dicts or dicts of dicts to create <server> elements for.

    Raises
    ------
    AssertionError
        * if username given without password or password without username
        * if no id given
    """

    if maven_servers is None:
        return

    assert isinstance(maven_servers, (dict, list))

    servers_element = ET.Element('servers')
    parent_element.append(servers_element)

    if isinstance(maven_servers, dict):
        for maven_server_key, maven_server_conf in maven_servers.items():
            assert \
                ( \
                    ('username' in maven_server_conf) \
                    and \
                    ('password' in maven_server_conf) \
                ) \
                or \
                ( \
                    ('username' not in maven_server_conf) \
                    and \
                    ('password' not in maven_server_conf) \
                ), \
                f"Configuration for maven server ({maven_server_key})" \
                f" must specify both 'username' and 'password' or neither: {maven_server_conf}"

            # if id is key in the dict then use that for the maven server id
            # else use the dict key for the maven server conf as the maven server id
            if 'id' in maven_server_conf:
                maven_server_id = maven_server_conf['id']
            else:
                maven_server_id = maven_server_key

            add_maven_server(
                parent_element=servers_element,
                maven_server_id=maven_server_id,
                maven_server_username=maven_server_conf.get('username'),
                maven_server_password=maven_server_conf.get('password')
            )
    elif isinstance(maven_servers, list):
        for maven_server_conf in maven_servers:
            assert 'id' in maven_server_conf, \
                "Configuration for maven servers" \
                f" must specify a 'id': {maven_server_conf}"

            assert \
                ( \
                    ('username' in maven_server_conf) \
                    and \
                    ('password' in maven_server_conf) \
                ) \
                or \
                ( \
                    ('username' not in maven_server_conf) \
                    and \
                    ('password' not in maven_server_conf) \
                ), \
                "Configuration for maven servers" \
                f" must specify both 'username' and 'password' or neither: {maven_server_conf}"

            maven_server_id = maven_server_conf['id']

            add_maven_server(
                parent_element=servers_element,
                maven_server_id=maven_server_id,
                maven_server_username=maven_server_conf.get('username'),
                maven_server_password=maven_server_conf.get('password')
            )

def add_maven_server(
    parent_element,
    maven_server_id,
    maven_server_username=None,
    maven_server_password=None
):
    """Adds a <server> element to the given parent element.

    Parameters
    ----------
    parent_element : ET.Element
        The parent element to append the new server too.

    maven_server_id : str
        ID of the maven server to add.

    maven_server_username : str, optional
        Username for the maven server to add.
        Only used if maven_server_password also provided.

    maven_server_password : str, optional
        Password for the maven server to add.
        Only used if maven_server_username is provided.
    """
    server_element = ET.Element('server')
    parent_element.append(server_element)
    server_id = ET.SubElement(server_element, 'id')
    server_id.text = maven_server_id
    if maven_server_username and maven_server_password:
        server_username = ET.SubElement(server_element, 'username')
        server_username.text = maven_server_username
        server_password = ET.SubElement(server_element, 'password')
        server_password.text = maven_server_password

def add_maven_repositories(parent_element, maven_repositories): # pylint: disable=invalid-name
    """Adds <repositories> element to parent element and child <repository> element for each
    given maven repository.

    Parameters
    ----------
    parent_element : ET.Element
        Parent element to add the <repositories> element to.

    maven_repositories : (dict, list)
        List of dicts or dicts of dicts to create <repository> elements for.

    Raises
    ------
    AssertionError
        * id not specified
        * url not specified
    """

    if maven_repositories is None:
        return

    assert isinstance(maven_repositories, (dict, list))

    profiles_element = ET.Element('profiles')
    parent_element.append(profiles_element)
    profile_element = ET.Element('profile')
    profiles_element.append(profile_element)
    repositories_element = ET.Element('repositories')
    profile_element.append(repositories_element)

    if isinstance(maven_repositories, dict):
        for maven_repository_key, maven_repository_conf in maven_repositories.items():
            assert 'url' in maven_repository_conf, \
                f"Configuration for maven repository ({maven_repository_key})" \
                f" must specify a 'url': {maven_repository_conf}"

            # if id is key in the dict then use that for the maven repository id
            # else use the dict key for the maven server conf as the maven server id
            if 'id' in maven_repository_conf:
                repository_id = maven_repository_conf['id']
            else:
                repository_id = maven_repository_key

            repository_url = maven_repository_conf['url']
            releases_enabled = maven_repository_conf.get('releases')
            snapshots_enabled = maven_repository_conf.get('snapshots')

            add_maven_repository(
                parent_element=repositories_element,
                repository_id=repository_id,
                repository_url=repository_url,
                releases_enabled=releases_enabled,
                snapshots_enabled=snapshots_enabled
            )
    elif isinstance(maven_repositories, list):
        for maven_repository_conf in maven_repositories:
            assert 'id' in maven_repository_conf, \
                "Configuration for maven repository" \
                f" must specify a 'id': {maven_repository_conf}"

            assert 'url' in maven_repository_conf, \
                "Configuration for maven repository" \
                f" must specify a 'url': {maven_repository_conf}"

            repository_id = maven_repository_conf['id']
            repository_url = maven_repository_conf['url']
            releases_enabled = maven_repository_conf.get('releases')
            snapshots_enabled = maven_repository_conf.get('snapshots')

            add_maven_repository(
                parent_element=repositories_element,
                repository_id=repository_id,
                repository_url=repository_url,
                releases_enabled=releases_enabled,
                snapshots_enabled=snapshots_enabled
            )

def add_maven_repository(
    parent_element,
    repository_id,
    repository_url,
    releases_enabled=None,
    snapshots_enabled=None
):
    """Add a <repository> element to the given parent element

    Parameters
    ----------
    parent_element : ET.Element
        Parent elemtnt to add the <repository> element to.

    repository_id : str
        Value for the <id> element

    repository_url : str
        Value for the <url> element

    releases_enabled : (str, bool), optional
        Value for <releases><enabled>{snapshots_enabled}</enabled></releases>

    snapshots_enabled : (str, bool), optional
        Value for <snapshots><enabled>{snapshots_enabled}</enabled></snapshots>

    """
    repository_element = ET.Element('repository')
    parent_element.append(repository_element)

    repository_id_element = ET.SubElement(repository_element, 'id')
    repository_id_element.text = repository_id

    repository_url_element = ET.SubElement(repository_element, 'url')
    repository_url_element.text = repository_url

    if releases_enabled is not None:
        repository_releases = ET.SubElement(repository_element, 'releases')
        repository_releases_enabled = ET.SubElement(repository_releases, 'enabled')
        repository_releases_enabled.text = str(releases_enabled)

    if snapshots_enabled is not None:
        repository_snapshots = ET.SubElement(repository_element, 'snapshots')
        repository_snapshots_enabled = ET.SubElement(repository_snapshots, 'enabled')
        repository_snapshots_enabled.text = str(snapshots_enabled)

def add_maven_mirrors(parent_element, maven_mirrors):
    """Add <mirror> element for given maven mirror to <mirrors> element to given parent element.

    Parameters
    ----------
    parent_element : ET.Element
        Parent element to add <mirrors> element to.

    maven_mirrors : (dict, list)
        List of dicts or dicts of dicts of maven mirror configurations.

    Raises
    ------
    AssertionError
        * if maven mirror id not given
        * if maven mirror url not given
        * if maven mirror mirror-of not given
    """

    if maven_mirrors is None:
        return

    assert isinstance(maven_mirrors, (dict, list))

    mirrors_element = ET.Element('mirrors')
    parent_element.append(mirrors_element)

    if isinstance(maven_mirrors, dict):
        for maven_mirror_key, maven_mirror_conf in maven_mirrors.items():
            assert 'url' in maven_mirror_conf, \
                f"Configuration for maven mirror ({maven_mirror_key})" \
                f" must specify a 'url': {maven_mirror_conf}"
            assert 'mirror-of' in maven_mirror_conf, \
                f"Configuration for maven mirror ({maven_mirror_key})" \
                f" must specify a 'mirror-of': {maven_mirror_conf}"

            # if id is key in the dict then use that for the maven mirror id
            # else use the dict key for the maven mirror conf as the maven mirror id
            if 'id' in maven_mirror_conf:
                maven_mirror_id = maven_mirror_conf['id']
            else:
                maven_mirror_id = maven_mirror_key

            add_maven_mirror(
                parent_element=mirrors_element,
                maven_mirror_id=maven_mirror_id,
                maven_mirror_url=maven_mirror_conf['url'],
                maven_mirror_mirror_of=maven_mirror_conf['mirror-of']
            )
    elif isinstance(maven_mirrors, list):
        for maven_mirror_conf in maven_mirrors:
            assert 'id' in maven_mirror_conf, \
                "Configuration for maven mirror" \
                f" must specify a 'id': {maven_mirror_conf}"
            assert 'url' in maven_mirror_conf, \
                "Configuration for maven mirror" \
                f" must specify a 'url': {maven_mirror_conf}"
            assert 'mirror-of' in maven_mirror_conf, \
                "Configuration for maven mirror" \
                f" must specify a 'mirror-of': {maven_mirror_conf}"

            add_maven_mirror(
                parent_element=mirrors_element,
                maven_mirror_id=maven_mirror_conf['id'],
                maven_mirror_url=maven_mirror_conf['url'],
                maven_mirror_mirror_of=maven_mirror_conf['mirror-of']
            )

def add_maven_mirror(
    parent_element,
    maven_mirror_id,
    maven_mirror_url,
    maven_mirror_mirror_of
):
    """Adds a <mirror> element to given parent element.

    Parameters
    ----------
    parent_element : ElementTree.Element
        Parent element to add new <mirror> element to

    maven_mirror_id : str
        Value for <id> element.

    maven_mirror_url : str
        Value for <url> element.

    maven_mirror_mirror_of : str
        Value for <mirrorOf> element.

    """
    mirror_element = ET.Element('mirror')
    parent_element.append(mirror_element)

    mirror_id = ET.SubElement(mirror_element, 'id')
    mirror_id.text = maven_mirror_id

    mirror_url = ET.SubElement(mirror_element, 'url')
    mirror_url.text = maven_mirror_url

    mirror_mirror_of = ET.SubElement(mirror_element, 'mirrorOf')
    mirror_mirror_of.text = maven_mirror_mirror_of

def write_effective_pom(
    pom_file_path,
    output_path,
    profiles=None
):
    """Generates the effective pom for a given pom and writes it to a given directory

    Parameters
    ----------
    pom_file_path : str
        Path to pom file to render the effective pom for.
    output_path : str
        Path to write the effective pom to.
    profiles : list
        Maven profiles to use when generating the effective pom.

    See
    ---
    * https://maven.apache.org/plugins/maven-help-plugin/effective-pom-mojo.html

    Returns
    -------
    str
        Absolute path to the written effective pom generated from the given pom file path.

    Raises
    ------
    StepRunnerException
        If issue generating effective pom.
    """

    if not os.path.isabs(output_path):
        raise StepRunnerException(
            f"Given output path ({output_path}) is not absolute which will mean your output"
            f" file will actually end up being relative to the pom file ({pom_file_path}) rather"
            " than your expected root. Rather then handling this, just give this function an"
            " absolute path."
            " If you are a user seeing this, a programmer messed up somewhere, report an issue."
        )

    profiles_arguments = ""
    if profiles:
        if isinstance(profiles, str):
            profiles = [profiles]
        profiles_arguments = ['-P', f"{','.join(profiles)}"]

    try:
        sh.mvn( # pylint: disable=no-member
            'help:effective-pom',
            f'-f={pom_file_path}',
            f'-Doutput={output_path}',
            *profiles_arguments
        )
    except sh.ErrorReturnCode as error:
        raise StepRunnerException(
            f"Error generating effective pom for '{pom_file_path}' to '{output_path}': {error}"
        ) from error

    return output_path

def get_effective_pom(
    work_dir_path,
    pom_file,
    profiles
):
    """Writes the effective pom to a file if it does not already exist and returns the path.

    Parameters
    ----------
    work_dir_path : str
        Path to write the effective pom to if it does not already exist
    pom_file : str
        Path to pom file to create the effective pom for.
    profiles : [str]
        Profile(s) to use when generating the effective pom

    Returns
    -------
    str
        Path to the written effective pom generated from the 'pom-file' value.
    """
    effective_pom_path = os.path.join(work_dir_path, 'effective-pom.xml')

    if not os.path.exists(effective_pom_path):
        write_effective_pom(
            pom_file_path=pom_file,
            output_path=effective_pom_path,
            profiles=profiles
        )

    return effective_pom_path

def run_maven( #pylint: disable=too-many-arguments
    mvn_output_file_path,
    settings_file,
    pom_file,
    phases_and_goals,
    tls_verify=True,
    additional_arguments=None,
    profiles=None,
    no_transfer_progress=True
):
    """Runs maven using the given configuration.

    Parameters
    ----------
    mvn_output_file_path : str
        Path to file containing the maven stdout and stderr output.
    phases_and_goals : [str]
        List of maven phases and/or goals to execute.
    additional_arguments : [str]
        List of additional arguments to use.
    pom_file : str (path)
        pom used when executing maven.
    tls_verify : boolean
        Disables TLS Verification if set to False
    profiles : [str]
        List of maven profiles to use.
    no_transfer_progress : boolean
        `True` to suppress the transfer progress of packages maven downloads.
        `False` to have the transfer progress printed.\
        See https://maven.apache.org/ref/current/maven-embedder/cli.html
    settings_file : str (path)
        Maven settings file to use.

    Returns
    -------
    str
        Standard Out and Standard Error from running Maven.

    Raises
    ------
    StepRunnerException
        If maven returns a none 0 exit code.
    """

    if not isinstance(phases_and_goals, list):
        phases_and_goals = [phases_and_goals]

    # create profile argument
    profiles_arguments = ""
    if profiles:
        profiles_arguments = ['-P', f"{','.join(profiles)}"]

    # create no transfer progress argument
    no_transfer_progress_argument = None
    if no_transfer_progress:
        no_transfer_progress_argument = '--no-transfer-progress'

    # create tls arguments
    tls_arguments = []
    if not tls_verify:
        tls_arguments += [
            '-Dmaven.wagon.http.ssl.insecure=true',
            '-Dmaven.wagon.http.ssl.allowall=true',
            '-Dmaven.wagon.http.ssl.ignore.validity.dates=true',
        ]

    if not additional_arguments:
        additional_arguments = []

    # run maven
    try:
        with open(mvn_output_file_path, 'w', encoding='utf-8') as mvn_output_file:
            out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                sys.stdout,
                mvn_output_file
            ])
            err_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                sys.stderr,
                mvn_output_file
            ])

            sh.mvn( # pylint: disable=no-member
                *phases_and_goals,
                '-f', pom_file,
                '-s', settings_file,
                *profiles_arguments,
                no_transfer_progress_argument,
                *tls_arguments,
                *additional_arguments,
                _out=out_callback,
                _err=err_callback
            )
    except sh.ErrorReturnCode as error:
        raise StepRunnerException(
            f"Error running maven. {error}"
        ) from error

def get_maven_plugin_xml_element_path(plugin_name):
    """Create XML element path for a given maven plugin.

    Parameters
    ----------
    plugin_name : str
        Maven plugin name to get the XML element path for

    Returns
    -------
    str
        XML element path for the given maven plugin.
    """
    return f'./mvn:build/mvn:plugins/mvn:plugin[mvn:artifactId="{plugin_name}"]'

def get_plugin_configuration_values(
    plugin_name,
    configuration_key,
    work_dir_path,
    pom_file,
    profiles=None,
    phases_and_goals=None,
    require_phase_execution_config=False
): # pylint: disable=too-many-arguments
    """Gets the value(s) of a given configuration key for a given maven plugin.

    Will create an effective pom out of the given pom so as to be able to inherit configuration
    from parent poms.

    Will search:
    * executions for given phase, return all matches
    * if no matching executions will search plugin non execution specific configuration

    Paramters
    ---------
    plugin_name : str
        Name of the maven plugin to get the configuration value(s) for.
    configuration_key : str
        Configuration key of the maven plugin to get the configuration value(s) for.
    work_dir_path : str
        A working path to use to write the effective pom to for searching.
    pom_file : str
        Maven pom file to create the effective pom for to then search for the plugin configuration.
    profiles : [str]
        List of maven profiles to activate when creating the effective pom to search.
    phases_and_goals : [str]
        List of phases and goals to search for specific plugin executions for for the configuration
        value(s).
    require_phase_execution_config : bool
        True if the found configuration must be in a plugin execution matching one of the given
        phases. False if the found configuration can be a default configuration.

    Raises
    ------
    RuntimeError
        If the given plugin can not be found in the effective pom.

    Returns
    -------
    [str]
        List of configuration values found, or empty list if none found.
    """
    # get effective pom
    effective_pom_file = get_effective_pom(
        work_dir_path=work_dir_path,
        pom_file=pom_file,
        profiles=profiles
    )

    # ensure plugin enabled
    plugin_xml_element_path = get_maven_plugin_xml_element_path(plugin_name)
    plugin = get_xml_element_by_path(
        effective_pom_file,
        xpath=plugin_xml_element_path,
        default_namespace='mvn'
    )
    if plugin is None:
        raise RuntimeError(
            f"Expected maven plugin ({plugin_name}) not found in "
            f" effective pom for given pom ({pom_file})."
        )

    # look for phase and/or goal specific plugin configuration
    configuration_values = []
    if phases_and_goals:
        for phase_or_goal in phases_and_goals:
            # look for phase specific plugin configuration
            # EX:
            #   <plugin>
            #     <groupId>org.apache.maven.plugins</groupId>
            #     <artifactId>maven-surefire-plugin</artifactId>
            #     <version>${surefire-plugin.version}</version>
            #     <configuration>
            #       <reportsDirectory>
            #         ${project.build.directory}/surefire-reports-unit-test
            #       </reportsDirectory>
            #     </configuration>
            #     <executions>
            #       <execution>
            #         <id>integration-tests</id>
            #         <phase>integration-test</phase> <!--NOTE: Matches on this-->
            #         <goals>
            #           <goal>test</goal>
            #         </goals>
            #         <configuration>
            #           <skipTests>${skipITs}</skipTests>
            #           <reportsDirectory>
            #             ${project.build.directory}/surefire-reports-uat <!--NOTE: selects this-->
            #           </reportsDirectory>
            #           <systemProperties>
            #             <java.util.logging.manager>
            #               org.jboss.logmanager.LogManager
            #             </java.util.logging.manager>
            #           </systemProperties>
            #           <includes>
            #             <include>**/*IT.*</include>
            #           </includes>
            #         </configuration>
            #       </execution>
            #     </executions>
            #   </plugin>
            phase_config = get_xml_element_text_by_path(
                effective_pom_file,
                xpath=f'{plugin_xml_element_path}/mvn:executions/' \
                    f'mvn:execution[mvn:phase="{phase_or_goal}"]/' \
                    f'mvn:configuration/mvn:{configuration_key}',
                default_namespace='mvn',
                find_all=True
            )
            if isinstance(phase_config, list):
                configuration_values += phase_config

            # look for goals specific plugin configuration
            # EX:
            #   <plugin>
            #     <groupId>org.apache.maven.plugins</groupId>
            #     <artifactId>maven-failsafe-plugin</artifactId>
            #     <version>2.22.2</version>
            #     <executions>
            #       <execution>
            #         <goals>
            #           <goal>integration-test</goal> <!--NOTE: Matches on this-->
            #           <goal>verify</goal>
            #         </goals>
            #         <configuration>
            #           <reportsDirectory> <!--NOTE: selects this-->
            #             ${project.build.directory}/failsafe-reports-execution
            #           </reportsDirectory>
            #         </configuration>
            #       </execution>
            #     </executions>
            #   </plugin>
            goal_config = get_xml_element_text_by_path(
                effective_pom_file,
                xpath=f'{plugin_xml_element_path}/mvn:executions/mvn:execution/' \
                    f'*[mvn:goal="{phase_or_goal}"]/../mvn:configuration/mvn:{configuration_key}',
                default_namespace='mvn',
                find_all=True
            )
            if isinstance(goal_config, list):
                configuration_values += goal_config

    # if didn't find any phase specific configuration, look for default configuration
    if not require_phase_execution_config and not configuration_values:
        default_config = get_xml_element_text_by_path(
            effective_pom_file,
            xpath=f'{plugin_xml_element_path}/mvn:configuration/mvn:{configuration_key}',
            default_namespace='mvn',
            find_all=True
        )
        if isinstance(default_config, list):
            configuration_values += default_config

    # de dup results and return
    configuration_values = list(set(configuration_values))
    configuration_values.sort()
    return configuration_values

def get_plugin_configuration_absolute_path_values(
    plugin_name,
    configuration_key,
    work_dir_path,
    pom_file,
    profiles=None,
    phases_and_goals=None,
    require_phase_execution_config=False
): # pylint: disable=too-many-arguments
    """Gets the value(s) of a given configuration key for a given maven plugin and converts
    them to absolute paths (if they arn't already), if they were relative paths, assumes,
    relative to the given pom file.

    Will create an effective pom out of the given pom so as to be able to inherit configuration
    from parent poms.

    Will search:
    * executions for given phase, return all matches
    * if no matching executions will search plugin non execution specific configuration

    Paramters
    ---------
    plugin_name : str
        Name of the maven plugin to get the configuration value(s) for.
    configuration_key : str
        Configuration key of the maven plugin to get the configuration value(s) for.
    work_dir_path : str
        A working path to use to write the effective pom to for searching.
    pom_file : str
        Maven pom file to create the effective pom for to then search for the plugin configuration.
    profiles : [str]
        List of maven profiles to activate when creating the effective pom to search.
    phases_and_goals : [str]
        List of phases and goals to search for specific plugin executions for for the configuration
        value(s).
    require_phase_execution_config : bool
        True if the found configuration must be in a plugin execution matching one of the given
        phases. False if the found configuration can be a default configuration.

    Raises
    ------
    RuntimeError
        If the given plugin can not be found in the effective pom.

    Returns
    -------
    [str]
        List of configuration values found, or empty list if none found.
    """
    absolute_path_config_values = []

    # get the configuration values
    config_values = get_plugin_configuration_values(
        plugin_name=plugin_name,
        configuration_key=configuration_key,
        work_dir_path=work_dir_path,
        pom_file=pom_file,
        profiles=profiles,
        phases_and_goals=phases_and_goals,
        require_phase_execution_config=require_phase_execution_config
    )

    # transform that configuration into absolute paths for consistency
    if config_values:
        for config_value in config_values:
            # if absolute path use as is
            # else if relative path assume its relative to the pom and calc absolute path
            if os.path.isabs(config_value):
                absolute_path_config_values.append(config_value)
            else:
                absolute_path_config_values.append(os.path.join(
                    os.path.dirname(os.path.abspath(pom_file)),
                    config_value
                ))

    return absolute_path_config_values



"""Abstract class for config value decryptor implementers.
"""

from abc import ABC, abstractmethod

class ConfigValueDecryptor(ABC):
    """Abstract class for config value decryptor implementers.
    """

    @abstractmethod
    def can_decrypt(self, config_value):
        """Determine if a given config value can be decrypted by this decryptor.

        Parameters
        ----------
        config_value : ConfigValue
            Determine if this decryptor can decrypt this configuration value.

        Returns
        -------
        bool
            True if this ConfigValueDecryptor can decrypt the given ConfigValue
            False if this ConfigValueDecryptor can NOT decrypt the given ConfigValue.
        """

    @abstractmethod
    def decrypt(self, config_value):
        """Decrypt the value of the given ConfigValue.

        Parameters
        ----------
        config_value : ConfigValue
            Decrypt the value of this ConfigValue.

        Returns
        -------
        obj or None
            Decrypted value of the ConfigValue
            None if this decryptor can't decrypt the given ConfigValue
        """
